/*-
 * Copyright (C) @BABOLO  2000 Jul 30
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: ra-sum.c,v 1.13 2001/08/28 00:16:32 babolo Exp $ 
 */

#include <sys/syslimits.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <err.h>

#ifndef TCPCASHE_COUNT
#define TCPCASHE_COUNT "%r-/_Cashe_C"
#endif  TCPCASHE_COUNT
#ifndef TCPCASHE_ADDR
#define TCPCASHE_ADDR "%r-/_Cashe_A"
#endif  TCPCASHE_ADDR
#ifndef UNIQT_LOG
#define UNIQT_LOG "/tmp/babolo/uniqt.log"
#endif  UNIQT_LOG
#ifndef UNIT_FREE_COUNT
#define UNIT_FREE_COUNT 1024
#endif  UNIT_FREE_COUNT
#ifndef UNIT_FREE_ADDR
#define UNIT_FREE_ADDR 256
#endif  UNIT_FREE_ADDR
#define COUNT_MEMBERN 3

#define HOST_BITS_PER_INT 32                     /* borrowed from argus (Carnegie Mellon University) */
#define argtimeval timeval

#if defined(__sgi) || defined(HAVE_SOLARIS) || defined(ultrix) || defined(__osf__) || defined(linux) || defined(AIX)

#define timelocal mktime

#if defined(__sgi)
#include <bstring.h>
#include <ulimit.h>

#if _MIPS_SZLONG == 64
#undef argtimeval
#define argtimeval irix5_timeval
#endif

#undef TCPSTATES
struct ether_addr
{   u_char  octet[6];
};
#endif
#endif

#if defined(AIX)
#define _SUN
#include <sys/select.h>
#include <net/nh.h>
#endif

#define arg_uint8   u_char
#define arg_int8    char
#define arg_uint16  u_short
#define arg_int16   short

#if HOST_BITS_PER_INT == 32
#define arg_uint32  u_int
#define arg_int32   int
#else
#define arg_uint32  u_long
#define arg_int32   long
#endif                                       /* END borrowed from argus (Carnegie Mellon University) */

#include <cons_def.h>
#include <cons_out.h>
#include "bufread.h"
#include "traf_ex_record.h"

struct tree_node
{   u_int32_t mask;
    u_int32_t value;
    u_int32_t eq;
    u_int32_t neq;
};
struct registry
{   u_int64_t from;
    u_int64_t to;
};

const char months[][4] =
{"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
,"","","","","","","","","","","",""
};
const char weekdays[][4] = {"Sun","Mon","Tue","Wen","Thu","Fri","Sat"};
const char weekrs[][3] = {"","","","","","",""};
struct tree_node empty_tree;
struct tree_node *ip_tree;
u_int32_t len_tree;
u_int64_t *pcount, *bcount;
char *myname, *n_uniqt;
int flag_v, flag_i, flag_S;
struct registry ****ip_registered;
u_int64_t globalast;
                                                                                     /****************/
                                                                                     /* IP TCP cashe */
u_int32_t *tcpcashe_addr_m;                                                          /****************/
#define tcpcashe_addr(X,Y) (tcpcashe_addr_m[(X) * 256 + (Y)])
u_int64_t *tcpcashe_count_m;
#define tcpcashe_count(X,Y) (tcpcashe_count_m[(X) * COUNT_MEMBERN + (Y)])
#define free_addr ((u_int32_t*)tcpcashe_count_m)[0]
#define free_count ((u_int32_t*)tcpcashe_count_m)[1]
int tcpcashe_addr_file, tcpcashe_count_file;
off_t unit_free_addr, unit_free_count;
#define DELTA_addr (unit_free_addr * sizeof(u_int32_t) * 256)
#define DELTA_count (unit_free_count * sizeof(u_int64_t) * COUNT_MEMBERN)
u_int32_t count_max_mon, addr_max_mon;

int putchar_buf(char c, char *buf, size_t *cur, size_t max)
{   if  (*cur >= max) return(-1);
    buf[(*cur)++] = c;
    return(0);
}

int put_escaped(char c, char *buf, size_t *cur, size_t max)
{   if  (*cur >= max) return(-1);
    switch (c)
    {case '\0': buf[*cur++] = '\\'; return(-1);
     case 'n' : buf[*cur++] = '\n';           break;
     case 't' : buf[*cur++] = '\t';           break;
     case 'v' : buf[*cur++] = '\v';           break;
     case 'b' : buf[*cur++] = '\b';           break;
     case 'r' : buf[*cur++] = '\r';           break;
     case 'f' : buf[*cur++] = '\f';           break;
     case '\\': buf[*cur++] = '\\';           break;
     case '\'': buf[*cur++] = '\'';           break;
     case '\"': buf[*cur++] = '\"';           break;
     case '0' : buf[*cur++] = '\0';           break;
     default  :
        buf[*cur++] = '\\';
        if  (putchar_buf(c, buf, cur, max)) return(-1);
    }
    return(0);
}

int put_f( char *f, char end_char, char *buf, size_t cur, size_t max, u_int64_t eth)
/* converts name format to name         */
/* return 1 if converted with no errors */
/* return 0 if not converted            */
/* return -1 if converted with errors   */
{   int c;
    struct tm *tm;
    time_t tloc;
    size_t bg, en, i;
    char *tmp = f;

    time(&tloc);
    tm = gmtime(&tloc);
    if (!f) return(0);
    while ((c = *f++))
        if (c == end_char) break;
        else if (c == '\\')
        {   if(put_escaped(c = *(f++), buf, &cur, max))
            {   if (flag_v) fprintf( stderr, "%s: Illegal format string: '\\0' after '\\\\' %s\n"
                                   , myname, tmp
                                   );
                return(-1);
            };
        }else if (c == '%')
        {   if  (!(c = *(f++)))
            {   if  (putchar_buf('%', buf, &cur, max)) return(-1);
        ;   ;   if  (flag_v) fprintf( stderr, "%s: Illegal format string: '\\0' after '%%' in %s\n"
                                   , myname, tmp
                                   );
        ;   ;   return(-1);
        ;   };
        ;   if  (c == end_char)
            {   if  (putchar_buf('%', buf, &cur, max)) return(-1);
        ;   ;   if  (flag_v)
                    fprintf( stderr, "%s: Illegal format string: end_char '%c' after '%%' in %s\n"
                           , myname, end_char, tmp
                           );
        ;   ;   return(-1);
        ;   };
            switch (c)
            {case '%':
                if  (putchar_buf('%', buf, &cur, max)) return(-1);
                break;
        ;   ;case 'y':
                if  (2 != snprintf(&buf[cur], max - cur, "%02u", (unsigned)(tm->tm_year % 100)))
                    return(-1);
                cur += 2;
                break;
        ;   ;case 'Y':
                if  (4 != snprintf(&buf[cur], max - cur, "%04u", (unsigned)(tm->tm_year + 1900)))
                    return(-1);
                cur += 4;
                break;
        ;   ;case 'M':
                if  (2 != snprintf(&buf[cur], max - cur, "%02u", (unsigned)(tm->tm_mon + 1)))
                    return(-1);
                cur += 2;
                break;
        ;   ;case 'l':
                if  (3 != snprintf(&buf[cur], max - cur, "%s", months[tm->tm_mon]))
                    return(-1);
                cur += 3;
                break;
        ;   ;case '':
                if  (3 != snprintf(&buf[cur], max - cur, "%s", months[tm->tm_mon + 12]))
                    return(-1);
                cur += 3;
                break;
        ;   ;case 'd':
                if  (2 != snprintf(&buf[cur], max - cur, "%02u", (unsigned)(tm->tm_mday)))
                    return(-1);
                cur += 2;
                break;
        ;   ;case 'D':
                if  (3 != snprintf(&buf[cur], max - cur, "%s", weekdays[tm->tm_wday]))
                    return(-1);
                cur += 3;
                break;
        ;   ;case '':
                if  (2 != snprintf(&buf[cur], max - cur, "%s", weekrs[tm->tm_wday]))
                    return(-1);
                cur += 2;
                break;
        ;   ;case 'h':
                if  (2 != snprintf(&buf[cur], max - cur, "%02u", (unsigned)(tm->tm_hour)))
                    return(-1);
                cur += 2;
                break;
        ;   ;case 'm':
                if  (2 != snprintf(&buf[cur], max - cur, "%02u", (unsigned)(tm->tm_min)))
                    return(-1);
                cur += 2;
                break;
        ;   ;case 's':
                if  (2 != snprintf(&buf[cur], max - cur, "%02u", (unsigned)(tm->tm_sec)))
                    return(-1);
                cur += 2;
                break;
        ;   ;case 'e':
                if  (14 != snprintf( &buf[cur], max - cur, "%02X%02X:%02X%02X:%02X%02X"
                                   , (u_int32_t)(eth >> 40) & 0xFF
                                   , (u_int32_t)(eth >> 32) & 0xFF
                                   , (u_int32_t)(eth >> 24) & 0xFF
                                   , (u_int32_t)(eth >> 16) & 0xFF
                                   , (u_int32_t)(eth >> 8) & 0xFF
                                   , (u_int32_t)eth & 0xFF
                    )              )
                    return(-1);
                cur += 14;
                break;
        ;   ;case 'E':
                if  (17 != snprintf( &buf[cur], max - cur, "%02X:%02X:%02X:%02X:%02X:%02X"
                                   , (u_int32_t)(eth >> 40) & 0xFF
                                   , (u_int32_t)(eth >> 32) & 0xFF
                                   , (u_int32_t)(eth >> 24) & 0xFF
                                   , (u_int32_t)(eth >> 16) & 0xFF
                                   , (u_int32_t)(eth >> 8) & 0xFF
                                   , (u_int32_t)eth & 0xFF
                    )              )
                    return(-1);
                cur += 17;
                break;
        ;   ;case 'r':
                bg = 0;
                en = strlen(n_uniqt);
                while ((c = *f++))
                    if (c == end_char) break;
                    else if (c == '-')
                    {   if  (en > 0)
                        {   en--;
                            while (en > 0 && n_uniqt[en] != '/') en--;
                            if  (en == 0 && putchar_buf('.', buf, &cur, max)) return(-1);
                        }else
                        {   if  (putchar_buf('/', buf, &cur, max)) return(-1);
                            if  (putchar_buf('.', buf, &cur, max)) return(-1);
                            if  (putchar_buf('.', buf, &cur, max)) return(-1);
                        }
                    }else
                    {   f--;
                        break;
                    }
                for (i = bg; i < en; i++) if (putchar_buf(n_uniqt[i], buf, &cur, max)) return(-1);
                break;
        ;   ;default :
                if  (putchar_buf('%', buf, &cur, max)) return(-1);
                if  (putchar_buf(c, buf, &cur, max)) return(-1);
        ;   }
        }else if (putchar_buf(c, buf, &cur, max)) return(-1);
    putchar_buf('\0', buf, &cur, max);
    return(1);
}

void uniqt(char *st)
{   int file = open(UNIQT_LOG, O_WRONLY | O_APPEND | O_EXLOCK);
    FILE *f = NULL;
    if  (file >= 0) f = fdopen(file, "ab");
    if  (f)
    {   time_t t = time(NULL);
        fprintf(f, "%s type %s file %s\n", ctime(&t), st, n_uniqt);
        fclose(f);
}   }

u_int64_t eatoq(struct ether_addr *ea)
{   int i;
    u_int64_t ll;
    for (i = 0, ll = 0; i < ETHER_ADDR_LEN; i++)
    {   ll <<= 8;
        ll |= ea->octet[i];
    }
    return(ll);
}

int set_bit(u_int32_t n, u_char *mask)
{   const u_char m[] = {0001, 0002, 0004, 0010, 0020, 0040, 0100, 0200};
    int i;
    i = (mask[n / 8] & m[n % 8]) ? 1 : 0;
    mask[n / 8] |= m[n % 8];
    return(i);
}

int test_bit(u_int32_t n, u_char *mask)
{   const u_char m[] = {0001, 0002, 0004, 0010, 0020, 0040, 0100, 0200};
    return((mask[n / 8] & m[n % 8]) ? 1 : 0);
}

void ip_registry(struct in_addr *inaddr, int from, int to)
{   u_char *ad;
    if  (flag_i)
    {   ad = (u_char*)inaddr;
        if (ip_registered == NULL)
            ip_registered = calloc((size_t)256, sizeof(struct registry***));
        if (ip_registered[ad[0]] == NULL)
            ip_registered[ad[0]] = calloc((size_t)256, sizeof(struct registry**));
        if (ip_registered[ad[0]][ad[1]] == NULL)
            ip_registered[ad[0]][ad[1]] = calloc((size_t)256, sizeof(struct registry*));
        if (ip_registered[ad[0]][ad[1]][ad[2]] == NULL)
            ip_registered[ad[0]][ad[1]][ad[2]] = calloc((size_t)256, sizeof(struct registry));
        ip_registered[ad[0]][ad[1]][ad[2]][ad[3]].from += from;
        ip_registered[ad[0]][ad[1]][ad[2]][ad[3]].to += to;
}   }

u_int32_t new_count()
{   u_int32_t l = free_count;
    if  (!l)
    {   struct stat sb;
        u_int32_t count_max, ll;
        if  (flag_v > 2) fprintf(stderr, "More tcpcashe_count\n");
        if  (fstat(tcpcashe_count_file, &sb) < 0)         err(EX_NOINPUT, "Can't stat tcpcashe_count");
        l = (u_int32_t)(sb.st_size / (sizeof(u_int64_t) * COUNT_MEMBERN));
        if  (munmap(tcpcashe_count_m, sb.st_size) < 0)  err(EX_NOINPUT, "Can't mmap tcpcashe_count_m");
        if  ((ftruncate(tcpcashe_count_file, DELTA_count + sb.st_size)) < 0)
                                                        err(EX_IOERR, "Can't truncate tcpcashe_count");
        if  (fstat(tcpcashe_count_file, &sb) < 0)         err(EX_NOINPUT, "Can't stat tcpcashe_count");
        if  (MAP_FAILED == (void*)(tcpcashe_count_m = mmap( NULL
                                                          , sb.st_size
                                                          , PROT_READ | PROT_WRITE
                                                          , MAP_SHARED
                                                          , tcpcashe_count_file
                                                          , 0
            )                     )                       )
                                                        err(EX_NOINPUT, "Can't mmap tcpcashe_count_m");
        count_max = (u_int32_t)(sb.st_size / (sizeof(u_int64_t) * COUNT_MEMBERN));
        for (ll = l; ll < count_max; ll++) tcpcashe_count(ll, 0) = (u_int64_t)ll + 1;
        tcpcashe_count(count_max - 1, 0) = 0;
        free_count = l;
    }
    if  (count_max_mon < l) count_max_mon = l;
    if  (l)
    {   free_count = tcpcashe_count(l, 0);
        tcpcashe_count(l, 0) = 0;
    }else
    {   fprintf(stderr, "No count\n"); /* XXXXX */
        exit(EX_SOFTWARE);
    }
    if  (flag_v > 3) fprintf( stderr, "NEW count %u\n", l);
    return(l);
}

u_int32_t new_addr()
{   u_int32_t l = free_addr;
    if  (!l)
    {   struct stat sb;
        u_int32_t addr_max, ll;
        if  (flag_v > 2) fprintf(stderr, "More tcpcashe_addr\n");
        if  (fstat(tcpcashe_addr_file, &sb) < 0)           err(EX_NOINPUT, "Can't stat tcpcashe_addr");
        l = (u_int32_t)(sb.st_size / (sizeof(u_int32_t) * 256));
        if  (munmap(tcpcashe_addr_m, sb.st_size) < 0)    err(EX_NOINPUT, "Can't mmap tcpcashe_addr_m");
        if  ((ftruncate(tcpcashe_addr_file, DELTA_addr + sb.st_size)) < 0)
                                                         err(EX_IOERR, "Can't truncate tcpcashe_addr");
        if  (fstat(tcpcashe_addr_file, &sb) < 0)           err(EX_NOINPUT, "Can't stat tcpcashe_addr");
        if  (MAP_FAILED == (void*)(tcpcashe_addr_m = mmap( NULL
                                                         , sb.st_size
                                                         , PROT_READ | PROT_WRITE
                                                         , MAP_SHARED
                                                         , tcpcashe_addr_file
                                                         , 0
            )                     )                      )
                                                         err(EX_NOINPUT, "Can't mmap tcpcashe_addr_m");
        addr_max = (u_int32_t)(sb.st_size / (sizeof(u_int32_t) * 256));
        for (ll = l; ll < addr_max; ll++) tcpcashe_addr(ll, 0) = (u_int32_t)ll + 1;
        tcpcashe_addr(addr_max - 1, 0) = 0;
        free_addr = l;
    }
    if  (addr_max_mon < l) addr_max_mon = l;
    if  (l)
    {   free_addr = tcpcashe_addr(l, 0);
        tcpcashe_addr(l, 0) = 0;
    }else
    {   fprintf(stderr, "No addr\n"); /* XXXXX */
        exit(EX_SOFTWARE);
    }
    if  (flag_v > 3) fprintf( stderr, "NEW addr %u\n", l);
    return(l);
}

u_int32_t get_counter_index( struct in_addr *src
                           , struct in_addr *dst
                           , arg_uint16 *sport
                           , arg_uint16 *dport
                           , struct argtimeval *tim
                           , int *reverse
                           )
{   int sy, dy, level;
    u_char sc[12], dc[12];
    u_int32_t si, di;
    u_int32_t result;
    u_int64_t qtime;
    qtime = ((u_int64_t)tim->tv_sec << 32) + (u_int64_t)tim->tv_usec * 67108864 / 15625;
    bcopy((u_char*)src, sc, 4); bcopy((u_char*)sport, &sc[4], 2);
    bcopy((u_char*)dst, dc, 4); bcopy((u_char*)dport, &dc[4], 2);
    bcopy(dc, &sc[6], 6);
    bcopy(sc, &dc[6], 6);
    if  (flag_v > 2) fprintf( stderr, "GET for src %u.%u.%u.%u:%u dst %u.%u.%u.%u:%u\n"
                            , sc[0], sc[1], sc[2], sc[3], sc[4] * 256 + sc[5]
                            , dc[0], dc[1], dc[2], dc[3], dc[4] * 256 + dc[5]
                            );
    sy = dy = 1;
    si = di = 0;
    *reverse = flag_S;
    for (level = 0; level < 12; level++)
    {   sy = sy && tcpcashe_addr(si, sc[level]);
        dy = dy && tcpcashe_addr(di, dc[level]);
        if (flag_S ? (!dy && sy) : (dy && !sy)) *reverse = !flag_S;
        if  (!(sy | dy))
        {   u_int32_t i;
            i = (level == 11) ? new_count() : new_addr();
            if  (*reverse)
            {   tcpcashe_addr(di, dc[level]) = i;
                di = i;
            }else
            {   tcpcashe_addr(si, sc[level]) = i;
                si = i;
            }
            if  (flag_v > 3) fprintf( stderr, "[%s%u(%u)]"
                                    , *reverse ? "r" : ""
                                    , i
                                    , *reverse ? dc[level] : sc[level]
                                    );
        }else
        {   if  (sy) si = tcpcashe_addr(si, sc[level]);
            if  (dy) di = tcpcashe_addr(di, dc[level]);
            if  (flag_v > 3)
            {   if  (!sy) fprintf( stderr, "[X(%u)-%u(%u)]", sc[level], di, dc[level]);
                else if  (!dy) fprintf( stderr, "[%u(%u)-X(%u)]", si, sc[level], dc[level]);
                else fprintf( stderr, "[%u-%u]", si, di);
    }   }   }
    if  (flag_v > 3) fprintf( stderr, "\n");
    if  (dy && sy)
    {   fprintf(stderr, "Symmetryc cashe\n");
        exit(EX_SOFTWARE);
    }
    result = *reverse ? di : si;
    tcpcashe_count(result, 2) = qtime;
    if  (globalast < qtime) globalast = qtime;
    return(result);
}

int del_counter_inner(int level, u_char *c, u_int32_t i)
{   int r = 1, k;
    if  (i && level > 11)
    {   for (k = COUNT_MEMBERN - 1; k > 0; k--) tcpcashe_count(i, k) = 0;
        tcpcashe_count(i, 0) = free_count;
        free_count = i;
    }else if (i && level)
    {   if  ((r = del_counter_inner(level + 1, c, tcpcashe_addr(i, c[level]))))
        {   tcpcashe_addr(i, c[level]) = 0;
            for (k = 0; k < 256 && r; k++) r = r && (tcpcashe_addr(i, k) == 0);
            if  (r)
            {   tcpcashe_addr(i, 0) = free_addr;
                free_addr = i;
        }   }
    }else if (i || level) r = 0;
    else if  ((r = del_counter_inner(level + 1, c, tcpcashe_addr(i, c[level]))))
        tcpcashe_addr(i, c[level]) = 0;
    if  (flag_v > 3) fprintf( stderr, "RET %s %u(%u) at L%d - %s\n"
                            , (level > 11) ? "count" : "addr"
                            , i, c[level], level
                            , (r) ? "succ" : "err"
                            );
    return(r);
}

void del_counter_index(struct in_addr *src, struct in_addr *dst, arg_uint16 *sport, arg_uint16 *dport)
{   u_char sc[12], dc[12];
    bcopy((u_char*)src, sc, 4); bcopy((u_char*)sport, &sc[4], 2);
    bcopy((u_char*)dst, dc, 4); bcopy((u_char*)dport, &dc[4], 2);
    bcopy(dc, &sc[6], 6);
    bcopy(sc, &dc[6], 6);
    if  (flag_v > 2) fprintf( stderr, "DEL for src %u.%u.%u.%u:%u dst %u.%u.%u.%u:%u\n"
                            , sc[0], sc[1], sc[2], sc[3], sc[4] * 256 + sc[5]
                            , dc[0], dc[1], dc[2], dc[3], dc[4] * 256 + dc[5]
                            );
    del_counter_inner(0, sc, (u_int32_t)0);
    del_counter_inner(0, dc, (u_int32_t)0);
}

void timeout_counter_index(int i0, int i1, int i2, int i3, int i4, int i5, int i6, int i7, int i8, int i9, int ia, int ib)
{   u_char sc[12], dc[12];
    sc[0] = i0; sc[1] = i1; sc[2] = i2; sc[3] = i3; sc[4] = i4; sc[5] = i5;
    dc[0] = i6; dc[1] = i7; dc[2] = i8; dc[3] = i9; dc[4] = ia; dc[5] = ib;
    bcopy(dc, &sc[6], 6);
    bcopy(sc, &dc[6], 6);
    if  (flag_v > 2) fprintf( stderr, "DEL for src %u.%u.%u.%u:%u dst %u.%u.%u.%u:%u\n"
                            , sc[0], sc[1], sc[2], sc[3], sc[4] * 256 + sc[5]
                            , dc[0], dc[1], dc[2], dc[3], dc[4] * 256 + dc[5]
                            );
    del_counter_inner(0, sc, (u_int32_t)0);
    del_counter_inner(0, dc, (u_int32_t)0);
}

int cache_true( int level
              , u_int32_t addr
              , u_int32_t addr_max, u_int32_t count_max
              , u_char *a_mask, u_char *c_mask
              )
{   int i, r = 1;
    if  (flag_v > 3) fprintf( stderr, "level %2d test addr=%u max=%u\n"
                            , level, addr
                            , ((level < 11) ? addr_max : count_max)
                            );
    if  (addr_max_mon < addr) addr_max_mon = addr;
    for (i = 0; r && i < 256; i++)
    {   if  (flag_v > 3)
        {   if  (tcpcashe_addr(addr, i)) fprintf( stderr, " %d %u\n", i, tcpcashe_addr(addr, i));
            else fprintf( stderr, " %d spare\n", i);
        }
        if  (tcpcashe_addr(addr, i))
        {   r &= (   tcpcashe_addr(addr, i) < ((level < 11) ? addr_max : count_max)
                 &&  !test_bit(tcpcashe_addr(addr, i), (level < 11) ? a_mask : c_mask)
                 );
            if  (level < 11)
            {   r &= cache_true( level + 1
                               , tcpcashe_addr(addr, i)
                               , addr_max, count_max
                               , a_mask, c_mask
                               );
            }else if  (count_max_mon < tcpcashe_addr(addr, i)) count_max_mon = tcpcashe_addr(addr, i);
            if  (flag_v && !r)
            {   fprintf( stderr, "Error at L%d: addr=%u, i=%d, tcpcashe_addr(addr, i)=%u %s\n"
                       , level, addr, i, tcpcashe_addr(addr, i)
                       ,   test_bit(tcpcashe_addr(addr, i), (level < 11) ? a_mask : c_mask)
                         ? "used"
                         : "free"
                       );
            }
            set_bit(tcpcashe_addr(addr, i), (level < 11) ? a_mask : c_mask);
    }   }
    return(r);
}

void do_sum( struct in_addr *inaddr
           , u_int32_t pdstcnt
           , u_int32_t psrccnt
           , u_int32_t bdstcnt
           , u_int32_t bsrccnt
           )
{   u_int32_t p;
    u_int32_t a = ntohl(inaddr->s_addr);
    char *b = "    ";
    if  (flag_v > 1) fprintf( stderr, " addr %u.%u.%u.%u%s\tps=%6u  pd=%6u  bs=%8u  bd=%8u\n"
                            , (a >> 24) & 0xFF, (a >> 16) & 0xFF, (a >> 8) & 0xFF, a & 0xFF
                            , &b[4 - ((((a >> 24) & 0xFF) < 10) + (((a >> 16) & 0xFF) < 10) + (((a >> 8) & 0xFF) < 10) + ((a & 0xFF) < 10))]
                            , psrccnt, pdstcnt, bsrccnt, bdstcnt
                            );
    /* hack anti argus error(?) */
    if ((bsrccnt & 0xFF000000) == 0xFF000000) bsrccnt = 0;
    if ((bdstcnt & 0xFF000000) == 0xFF000000) bdstcnt = 0;
    /* End of hack */
    for (p = 0; ip_tree[p].mask != 0;)
    {   if  (flag_v > 3)
            fprintf(stderr, "addr %08X maddr %08X v %08X\n", a, ip_tree[p].mask & a, ip_tree[p].value);
        if  ((ip_tree[p].mask & a) == ip_tree[p].value) p = ip_tree[p].eq;
        else p = ip_tree[p].neq;
    }
    if  (flag_v > 2) fprintf(stderr, "categ %u\n", ip_tree[p].eq);
    pcount[ip_tree[p].eq] += psrccnt;
    bcount[ip_tree[p].eq] += bsrccnt;
    if  (flag_v > 2) fprintf(stderr, "categ %u\n", ip_tree[p].neq);
    pcount[ip_tree[p].neq] += pdstcnt;
    bcount[ip_tree[p].neq] += bdstcnt;
}

void print_ip(FILE *f, u_int32_t ip)
{   fprintf(f, "%3u", (ip >> 24) & 0xFF);
    fprintf(f, ".%u", (ip >> 16) & 0xFF);
    fprintf(f, ".%u", (ip >> 8) & 0xFF);
    fprintf(f, ".%u", ip & 0xFF);
}

u_int64_t readether(char *in)
/* reads in as MAC address and converts to u_int_64 */
{   u_int32_t i = 0, k = 0;
    char *p, run = 1;
    u_int64_t eth = 0, j = 0;
    for (p = in; p && run; p++)
    {   switch (*p)
        {case '0':case '1':case '2':case '3':case '4':case '5':case '6':case '7':case '8':case '9':
            if  (++i > 12)
            {   fprintf(stderr, "%s: readether: ether addr longer then 6 bytes\n", myname);
                exit(EX_DATAERR);
            }
            k++;
            j <<=4;
            j |= *p - '0';
            break;
         case 'a':case 'b':case 'c':case 'd':case 'e':case 'f':
            if  (++i > 12)
            {   fprintf(stderr, "%s: readether: ether addr longer then 6 bytes\n", myname);
                exit(EX_DATAERR);
            }
            k++;
            j <<=4;
            j |= *p - 'a' + 10;
            break;
         case 'A':case 'B':case 'C':case 'D':case 'E':case 'F':
            if  (++i > 12)
            {   fprintf(stderr, "%s: readether: ether addr longer then 6 bytes\n", myname);
                exit(EX_DATAERR);
            }
            k++;
            j <<=4;
            j |= *p - 'A' + 10;
            break;
         case '\0':
            run = 0;
         case ':':
            run++;
         default:
            run--;
            i += k & 1;
            k += k & 1;
            eth <<= k * 4;
            eth |= j;
            k = 0;
            j = 0;
    }   }
    if  (i != 12)
    {   fprintf(stderr, "%s: readether: ether addr must be 6 bytes instead of %s~\n", myname, in);
        exit(EX_DATAERR);
    }
    return(eth);
}

void usage(int status)
{   fprintf( stderr
           , "ra-sum @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "ra-sum [-c policy] -e etheraddr -r input [-p pluspacket] [-C Ccashe] [-A Acashe]\n"
             "  -t traffic\n"
             "  -i IP addresses\n"
             "  -g look at cashe\n"
             "  -k[k[k]] delete or truncate cashe\n"
             "  -K not detailed mode of argus\n"
             "  -f force free memory cashe test\n"
           );
    exit(status);
}

int main(int argc, char **argv)
{   int c;
    int fin, file;
    struct stat sb;
    u_int64_t eth;
    struct WriteStruct next;
    int pluspacket;
    u_int32_t l, c_max, count_max, addr_max;
    int flag_t, flag_g, flag_b, flag_k, flag_f, new_cashe, no_cashe;
    int trunk_a, trunk_c;
    u_int64_t n, timeout;
    char tcpcashe_addr_name[PATH_MAX + 8], tcpcashe_count_name[PATH_MAX + 8];
    char *tcpcashe_addr_f, *tcpcashe_count_f, *report_f;
    u_char *a_mask, *c_mask;
    struct traf_ex_record out;

    myname = *argv;
    flag_v = flag_t = flag_i = flag_g = flag_b = flag_S = flag_k = flag_f = no_cashe = 0;
    bzero((void*)&empty_tree, sizeof(empty_tree));
    bzero(tcpcashe_count_name, PATH_MAX + 8);
    bzero(tcpcashe_addr_name, PATH_MAX + 8);
    tcpcashe_count_f = TCPCASHE_COUNT;
    tcpcashe_addr_f = TCPCASHE_ADDR;
    unit_free_count = UNIT_FREE_COUNT;
    unit_free_addr = UNIT_FREE_ADDR;
    ip_tree = &empty_tree;
    ip_registered = NULL;
    fin = fileno(stdin);
    pluspacket = 42;
    report_f = NULL;
    eth = 0; eth--;
    timeout = 2000;
    globalast = 0;
    new_cashe = 0;
    len_tree = 1;
    n_uniqt = "";
    while ((c = getopt(argc, argv, "vSbtigkKfr:e:c:p:C:A:T:h")) > 0)
    {   switch (c)
        {case 'v':
            flag_v++;
            break;
         case 'S':
            flag_S++;
            break;
         case 'b':
            flag_b++;
            break;
         case 't':
            flag_t++;
            break;
         case 'i':
            flag_i++;
            break;
         case 'g':
            flag_g++;
            break;
         case 'k':
            flag_k++;
            break;
         case 'K':
            no_cashe++;
            break;
         case 'f':
            flag_f++;
            break;
         case 'r':
            n_uniqt = optarg;
            if  (!(fin = open(optarg, O_RDONLY)))             err(EX_NOINPUT, "Can't open %s", optarg);
            break;
         case 'e':
            eth = readether(optarg);
            break;
         case 'c':
            if  ((file = open(optarg, O_RDONLY)) < 0)         err(EX_NOINPUT, "Can't open %s", optarg);
            if  (fstat(file, &sb) < 0)                        err(EX_NOINPUT, "Can't stat %s", optarg);
            ip_tree = (struct tree_node*)mmap( NULL
                                             , sb.st_size
                                             , PROT_READ
                                             , 0
                                             , file
                                             , 0
                                             );
            if  ((void*)ip_tree == MAP_FAILED)                err(EX_NOINPUT, "Can't mmap %s", optarg);
            len_tree = sb.st_size / sizeof(struct tree_node);
            close(file);
            break;
         case 'p':
            pluspacket = strtol(optarg, NULL, 0);
            break;
         case 'C':
            tcpcashe_count_f = optarg;
            break;
         case 'A':
            tcpcashe_addr_f = optarg;
            break;
         case 'T':
            timeout = strtoq(optarg, NULL, 0);
            break;
         case 'h': usage(EX_OK);
         default: usage(EX_USAGE);
    }   }
    timeout <<= 32;
    if  (eth == ~0 && !flag_g)
    {   fprintf(stderr, "%s: exact one ether addr must be given\n", myname);          exit(EX_DATAERR);
    }
    if  (flag_i && flag_t && flag_b)
    {   fprintf(stderr, "%s: can't use -itb flags together\n", myname);               exit(EX_DATAERR);
    }
    if  (0 >= put_f(tcpcashe_count_f, '\0', tcpcashe_count_name, 0, PATH_MAX, eth))
    {   fprintf(stderr, "%s: tcpcashe_count name format wrong\n", myname);
        exit(EX_DATAERR);
    }
    if  (0 >= put_f(tcpcashe_addr_f, '\0', tcpcashe_addr_name, 0, PATH_MAX, eth))
    {   fprintf(stderr, "%s: tcpcashe_addr name format wrong\n", myname);
        exit(EX_DATAERR);
    }

    argc -= optind;
    argv += optind;

                                                  /* how many account planes for in and out traffic? */
    if (flag_v > 1) fprintf(stderr, "Accounting policy:\n");
    for (l = 0, c_max = 0; l < len_tree; l++)
    {   if  (ip_tree[l].mask == 0)
        {   if  (ip_tree[l].eq > c_max) c_max = ip_tree[l].eq;
            if  (ip_tree[l].neq > c_max) c_max = ip_tree[l].neq;
            if  (flag_v > 1) fprintf(stderr, "%6u  %6uIn, %6uOut\n", l, ip_tree[l].eq, ip_tree[l].neq);
        }else if (flag_v > 1)
        {   fprintf(stderr, "%6u ", l);
            print_ip(stderr, ip_tree[l].value);
            fprintf(stderr, "\t       ");
            print_ip(stderr, ip_tree[l].mask);
            fprintf(stderr, "\t%6u= %6u<>\n", ip_tree[l].eq, ip_tree[l].neq);
    }   }
    c_max++;

    if  (flag_v) fprintf ( stderr, "%02X%02X:%02X%02X:%02X%02X c=%u\naddr =%s~\ncount=%s~\n"
                         , (u_int32_t)(eth >> 40) & 0xFF
                         , (u_int32_t)(eth >> 32) & 0xFF
                         , (u_int32_t)(eth >> 24) & 0xFF
                         , (u_int32_t)(eth >> 16) & 0xFF
                         , (u_int32_t)(eth >> 8) & 0xFF
                         , (u_int32_t)eth & 0xFF
                         , c_max
                         , tcpcashe_addr_name
                         , tcpcashe_count_name
                         );
                                                               /* global counters for account planes */
    if  (flag_t)
    {   pcount = (u_int64_t*)malloc(sizeof(u_int64_t) * c_max);
        bcount = (u_int64_t*)malloc(sizeof(u_int64_t) * c_max);
        if (pcount == NULL || bcount == NULL)                                     err(EX_CONFIG, NULL);
    }

                                                                                   /* counters cashe */
    tcpcashe_count_file = open(tcpcashe_count_name, O_RDWR | O_EXLOCK);
    if  (tcpcashe_count_file < 0)
    {   if  ((tcpcashe_count_file = open(tcpcashe_count_name, O_RDWR | O_EXLOCK | O_CREAT, 0666)) <= 0)
                                             err(EX_CANTCREAT, "Can't create %s", tcpcashe_count_name);
        if  (flag_v > 2) fprintf(stderr, "New tcpcashe_count_file = %d\n", tcpcashe_count_file);
        new_cashe = 1;
    }else if  (flag_v > 2) fprintf(stderr, "tcpcashe_count_file = %d\n", tcpcashe_count_file);

                                                                                       /* addr cashe */
    tcpcashe_addr_file = open(tcpcashe_addr_name, O_RDWR | O_EXLOCK);
    if  (tcpcashe_addr_file < 0)
    {   if  ((tcpcashe_addr_file = open(tcpcashe_addr_name, O_RDWR | O_EXLOCK | O_CREAT, 0666)) <= 0)
                                              err(EX_CANTCREAT, "Can't create %s", tcpcashe_addr_name);
        if  (flag_v > 2) fprintf(stderr, "New tcpcashe_addr_file = %d\n", tcpcashe_addr_file);
        new_cashe = 1;
    }else if  (flag_v > 2) fprintf(stderr, "tcpcashe_addr_file = %d\n", tcpcashe_addr_file);

    if  (new_cashe)
    {                                                                         /* init counters cashe */
    ;   if  (ftruncate(tcpcashe_count_file, 0) < 0)
                                               err(EX_IOERR, "Can't truncate %s", tcpcashe_count_name);
    ;   if  (ftruncate(tcpcashe_count_file, DELTA_count) < 0)
                                               err(EX_IOERR, "Can't truncate %s", tcpcashe_count_name);
    ;   if  (fstat(tcpcashe_count_file, &sb) < 0) err(EX_NOINPUT, "Can't stat %s", tcpcashe_count_name);
    ;   if  (MAP_FAILED == (void*)(tcpcashe_count_m = mmap( NULL
                                                          , sb.st_size
                                                          , PROT_READ | PROT_WRITE
                                                          , MAP_SHARED
                                                          , tcpcashe_count_file
                                                          , 0
            )                     )                       )
                                                          err(EX_NOINPUT, "Can't mmap tcpcashe_count");
    ;   count_max = (u_int32_t)(sb.st_size / (sizeof(u_int64_t) * COUNT_MEMBERN));
    ;   for (l = 1; l < count_max; l++) tcpcashe_count(l, 0) = (u_int64_t)l + 1;
    ;   tcpcashe_count(unit_free_count - 1, 0) = 0;
    ;   free_count = 1;

                                                                                 /* init addrs cashe */
    ;   if  ((ftruncate(tcpcashe_addr_file, 0)) < 0)
                                                err(EX_IOERR, "Can't truncate %s", tcpcashe_addr_name);
    ;   if  (flag_v > 2) fprintf(stderr, "Clear tcpcashe_addr\n");
    ;   if  ((ftruncate(tcpcashe_addr_file, DELTA_addr)) < 0)
                                                err(EX_IOERR, "Can't truncate %s", tcpcashe_addr_name);
    ;   if  (fstat(tcpcashe_addr_file, &sb) < 0)  err(EX_NOINPUT, "Can't stat %s", tcpcashe_addr_name);
    ;   if  (MAP_FAILED == (void*)(tcpcashe_addr_m = mmap( NULL
                                                         , sb.st_size
                                                         , PROT_READ | PROT_WRITE
                                                         , MAP_SHARED
                                                         , tcpcashe_addr_file
                                                         , 0
            )                     )                      )
                                                         err(EX_NOINPUT, "Can't mmap tcpcashe_addr_m");
    ;   addr_max = (u_int32_t)(sb.st_size / (sizeof(u_int32_t) * 256));
    ;   for (l = 1; l < addr_max; l++) tcpcashe_addr(l, 0) = (u_int32_t)l + 1;
    ;   tcpcashe_addr(unit_free_addr - 1, 0) = 0;
    ;   free_addr = 1;
    }else
    {   if  (fstat(tcpcashe_count_file, &sb) < 0) err(EX_NOINPUT, "Can't stat %s", tcpcashe_count_name);
    ;   if  (MAP_FAILED == (void*)(tcpcashe_count_m = mmap( NULL
                                                          , sb.st_size
                                                          , PROT_READ | PROT_WRITE
                                                          , MAP_SHARED
                                                          , tcpcashe_count_file
                                                          , 0
            )                     )                       )
                                                          err(EX_NOINPUT, "Can't mmap tcpcashe_count");
    ;   count_max = (u_int32_t)(sb.st_size / (sizeof(u_int64_t) * COUNT_MEMBERN));
    ;   if  (fstat(tcpcashe_addr_file, &sb) < 0)  err(EX_NOINPUT, "Can't stat %s", tcpcashe_addr_name);
    ;   if  (MAP_FAILED == (void*)(tcpcashe_addr_m = mmap( NULL
                                                         , sb.st_size
                                                         , PROT_READ | PROT_WRITE
                                                         , MAP_SHARED
                                                         , tcpcashe_addr_file
                                                         , 0
            )                     )                      )
                                                         err(EX_NOINPUT, "Can't mmap tcpcashe_addr_m");
    ;   addr_max = (u_int32_t)(sb.st_size / (sizeof(u_int32_t) * 256));
    }

    if  (flag_v) fprintf(stderr, "free_addr = %u\n", free_addr);
    if  (flag_v) fprintf(stderr, "free_count = %u\n", free_count);
    count_max_mon = free_count;
    addr_max_mon = free_addr;

                                                                                       /* view cashe */
    if  (flag_g)
    {   int       i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, ib;
        u_int32_t t0, t1, t2, t3, t4, t5, t6, t7, t8, t9, ta, tb;
        struct tm *tm;
        for (i0 = 0; i0 < 256; i0++)
        { if ((t0 = tcpcashe_addr(0, i0)))
          { for (i1 = 0; i1 < 256; i1++)
            { if ((t1 = tcpcashe_addr(t0, i1)))
              { for (i2 = 0; i2 < 256; i2++)
                { if ((t2 = tcpcashe_addr(t1, i2)))
                  { for (i3 = 0; i3 < 256; i3++)
                    { if ((t3 = tcpcashe_addr(t2, i3)))
                      { for (i4 = 0; i4 < 256; i4++)
                        { if ((t4 = tcpcashe_addr(t3, i4)))
                          { for (i5 = 0; i5 < 256; i5++)
                            { if ((t5 = tcpcashe_addr(t4, i5)))
                              { for (i6 = 0; i6 < 256; i6++)
                                { if ((t6 = tcpcashe_addr(t5, i6)))
                                  { for (i7 = 0; i7 < 256; i7++)
                                    { if ((t7 = tcpcashe_addr(t6, i7)))
                                      { for (i8 = 0; i8 < 256; i8++)
                                        { if ((t8 = tcpcashe_addr(t7, i8)))
                                          { for (i9 = 0; i9 < 256; i9++)
                                            { if ((t9 = tcpcashe_addr(t8, i9)))
                                              { for (ia = 0; ia < 256; ia++)
                                                { if ((ta = tcpcashe_addr(t9, ia)))
                                                  { for (ib = 0; ib < 256; ib++)
                                                    { if ((tb = tcpcashe_addr(ta, ib)))
                                                      { tm = localtime((time_t*)&tcpcashe_count(tb, 2) + 1);
                                                        fprintf( stderr
                                                               , " %4d-%s-%02d %02d:%02d:%02d.%03u"
                                                                 " src %u.%u.%u.%u:%u"
                                                                 " dst %u.%u.%u.%u:%u"
                                                                 " %20qu %20qu\n"
                                                               , 1900 + tm->tm_year
                                                               , months[tm->tm_mon]
                                                               , tm->tm_mday
                                                               , tm->tm_hour, tm->tm_min, tm->tm_sec
                                                               , (unsigned)((tcpcashe_count(tb, 2) & 0xffffffff) * 125 / 536870912)
                                                               , i0, i1, i2, i3, i4 * 256 + i5
                                                               , i6, i7, i8, i9, ia * 256 + ib
                                                               , tcpcashe_count(tb, 0)
                                                               , tcpcashe_count(tb, 1)
                                                               );
    }   } } } } } } } } } } } } } } } } } } } } } } } }

                                                                             /************************/
                                                                             /* test cashe integrity */
    a_mask = (u_char*)calloc((addr_max + 7) / 8, 1);                         /************************/
    c_mask = (u_char*)calloc((count_max + 7) / 8, 1);
    if  (a_mask == NULL || c_mask == NULL)                     err(EX_UNAVAILABLE, "No mem for masks");
                                                                                        /* mark used */
    trunk_a = trunk_c = 0;
    if  (flag_f || flag_k)
    {   if  (!cache_true( 0, 0, addr_max, count_max, a_mask, c_mask))
        {   fprintf(stderr, "cashe is broken\n");                                     exit(EX_DATAERR);
    ;   }
    ;   if  (flag_v)
        {   for (l = 0, n = 0; l < addr_max; l++)
                if  (test_bit(l, a_mask))
                {   n++;
    ;               if  (flag_v > 2) fprintf(stderr, " %u", l);
    ;           }
    ;       if  (flag_v > 2) fprintf(stderr, " A\n");
    ;       fprintf(stderr, "%qu used addr blocks\n", ++n);
    ;       for (l = 0, n = 0; l < count_max; l++)
                if  (test_bit(l, c_mask))
                {   n++;
    ;               if  (flag_v > 2) fprintf(stderr, " %u", l);
    ;           }
    ;       if  (flag_v > 2) fprintf(stderr, " C\n");
    ;       fprintf(stderr, "%qu used counters\n", n);
    ;   }
    ;   if  (flag_k)
        {   u_int32_t ac = 0, cc = 0, as, cs, at = 1, ct = 1;
    ;       for (l = 1; l < addr_max; l++)
            {   if  (!test_bit(l, a_mask))
                {   ac++;
    ;               at = 1;
    ;           }else if (at)
                {   as = l;
    ;               at = 0;
    ;       }   }
    ;       for (l = 1; l < count_max; l++)
            {   if  (!test_bit(l, c_mask))
                {   cc++;
    ;               ct = 1;
    ;           }else if (ct)
                {   cs = l;
    ;               ct = 0;
    ;       }   }
    ;       if  ((addr_max / ((flag_k == 1) ? 8 : 2)) > (ac + unit_free_addr)) trunk_a = 1;
    ;       if  ((count_max / ((flag_k == 1) ? 8 : 2)) > (ac + unit_free_count)) trunk_c = 1;
    }   }
                                                                     /* mark spare (in list of free) */
    if  (flag_f || trunk_a)
    {   u_int32_t *al;
    ;   for (al = &free_addr, l = *al; l;)
        {   l = *al;
    ;       if  (flag_v > 3) fprintf(stderr, "Free addr %u\n", l);
    ;       if  (l >= addr_max)
            {   fprintf(stderr, "free addr list of cashe is broken: %d >= %d\n", l, addr_max);
            ;                                                                         exit(EX_DATAERR);
    ;       }
    ;       if  (set_bit(l, a_mask))
            {   fprintf(stderr, "free addr list of cashe is broken: %d not free\n", l);
            ;                                                                         exit(EX_DATAERR);
    ;       }
    ;       if  (l > tcpcashe_addr(l, 0))
            {   *al = tcpcashe_addr(l, 0);
    ;           al = &tcpcashe_addr(*al, 0);
    ;           tcpcashe_addr(l, 0) = tcpcashe_addr(*al, 0);
    ;           tcpcashe_addr(*al, 0) = l;
    ;       }else
            {   al = &tcpcashe_addr(l, 0);
    ;   }   }
    ;   for (l = 1; l < addr_max; l++)
            if (!test_bit(l, a_mask)) fprintf(stderr, "%u addr block lost\n", l);
    }

    if  (flag_f || trunk_c)
    {   for (l = free_count; l; l = tcpcashe_count(l, 0))
        {   if  (flag_v > 3) fprintf(stderr, "Free counter %u\n", l);
    ;       if  (l >= count_max)
            {   fprintf(stderr, "free count list of cashe is broken: %d >= %d\n", l, count_max);
            ;                                                                         exit(EX_DATAERR);
            }
    ;       if  (set_bit(l, c_mask))
            {   fprintf(stderr, "free count list of cashe is broken: %d not free\n", l);
            ;                                                                         exit(EX_DATAERR);
    ;   }   }
    ;   for (l = 1; l < count_max; l++)
            if (!test_bit(l, c_mask)) fprintf(stderr, "%u counter lost\n", l);
    }

    free(a_mask);
    free(c_mask);

    init_next();                                                                        /* main loop */
    for (n = 0; (flag_i || flag_t) && get_next(fin, &next) == sizeof(next); n++)
    {   u_int32_t status = ntohl(next.status);
        char *mc;
        mc = " ";
    ;   if  (n == 0)
        {   if  (status & ARGUSCONTROL);
            else
            {   fprintf(stderr, "data error\n");                                      exit(EX_DATAERR);
    ;       }
    ;   }else if (  status & ARGUSCONTROL);
        else if  (  status & IPPROTO
        /*       || status & ARPPROTO */  /* XXXXX */
                 )
        {   u_int64_t srccnt, dstcnt, srcby, dstby;
            next.ws_ip.startime.tv_sec  = ntohl(next.ws_ip.startime.tv_sec);
            next.ws_ip.startime.tv_usec = ntohl(next.ws_ip.startime.tv_usec);
            next.ws_ip.lasttime.tv_usec = ntohl(next.ws_ip.lasttime.tv_usec);
            ((arg_uint16 *)&next.ws_ip.lasttime.tv_sec)[0] = 0;
            next.ws_ip.lasttime.tv_sec  = ntohl(next.ws_ip.lasttime.tv_sec);
            next.ws_ip.lasttime.tv_sec += next.ws_ip.startime.tv_sec;

    ;       if  (!flag_t && !flag_i);
            else if (status & ICMPPROTO)
            {   srccnt = 0;
    ;       ;   dstcnt = 0;
    ;       ;   if  (status & CON_ESTABLISHED)
                {   srccnt = ntohl(next.ws_ip_udp.src_count);
    ;       ;       dstcnt = ntohl(next.ws_ip_udp.dst_count);
    ;       ;   }else if  (   (   (status & TIMED_OUT)
                              ?   ((unsigned char *) &next.ws_ip_udp.src_bytes)[0]
                              :   next.ws_ip_icmp.type
                              )
                          ==  ICMP_ECHOREPLY
                          ) dstcnt = 1;
                else  srccnt = 1;
    ;       ;   if  (eatoq(&next.ws_ip_phys.ethersrc) == eth)
                {   if  (status & REVERSE)
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu icmp srcR%s", n, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_dst, dstcnt, srccnt, pluspacket * dstcnt, pluspacket * srccnt);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_src, dstcnt, srccnt);
    ;       ;       }else
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu icmp src %s", n, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_dst, srccnt, dstcnt, pluspacket * srccnt, pluspacket * dstcnt);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_src, srccnt, dstcnt);
    ;       ;   }   }
    ;       ;   if  (eatoq(&next.ws_ip_phys.etherdst) == eth)
                {   if  (status & REVERSE)
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu icmp dstR%s", n, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_src, srccnt, dstcnt, pluspacket * srccnt, pluspacket * dstcnt);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_dst, srccnt, dstcnt);
    ;       ;       }else
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu icmp dst %s", n, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_src, dstcnt, srccnt, pluspacket * dstcnt, pluspacket * srccnt);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_dst, dstcnt, srccnt);
    ;       ;   }   }
    ;       }else if  (  status & (UDPPROTO | TCPPROTO))
            {   char *proto = "ip ";
    ;       ;   srccnt = ntohl(next.ws_ip_udp.src_count);
    ;       ;   dstcnt = ntohl(next.ws_ip_udp.dst_count);
    ;       ;   srcby = dstby = 0;
    ;       ;   if  (status & (CON_ESTABLISHED | UDPPROTO) ? 1 : !(status & (SAW_SYN | SAW_SYN_SENT)))
                {   srcby = ntohl(next.ws_ip_udp.src_bytes);
    ;       ;       dstby = ntohl(next.ws_ip_udp.dst_bytes);
    ;       ;   }
    ;       ;   if  (!(status & (CON_ESTABLISHED | SAW_SYN | SAW_SYN_SENT)))
                {   mc = "N";
    ;       ;       uniqt("ip   xxxN");
    ;       ;   }
    ;       ;   srcby += pluspacket * srccnt;
    ;       ;   dstby += pluspacket * dstcnt;
    ;       ;   if  (status & TCPPROTO && flag_t)
                {   int reverse;
                    u_int32_t l;
    ;       ;       proto = "tcp";
    ;       ;       l = get_counter_index( &next.ws_ip_src
                                         , &next.ws_ip_dst
                                         , &next.ws_ip.sport
                                         , &next.ws_ip.dport
                                         , &next.ws_ip.lasttime
                                         , &reverse
                                         );
    ;       ;       if  (status & (NORMAL_CLOSE | TIMED_OUT | RESET))
                    {   if  (!no_cashe)
                        {   if  (srcby < tcpcashe_count(l, reverse ? 1 : 0))
                            {   if  (  flag_v > 1
                                    && pluspacket * 3 + srcby < tcpcashe_count(l, reverse ? 1 : 0)
                                    )
                                    fprintf( stderr, "In file=%s record=%qu status=%08X %qu < %qu\n"
                                           , n_uniqt, n, status
                                           , srcby, tcpcashe_count(l, reverse ? 1 : 0)
                                           );
            ;                   srcby = 0;
    ;       ;               }else srcby -= tcpcashe_count(l, reverse ? 1 : 0);

    ;       ;               if  (dstby < tcpcashe_count(l, reverse ? 0 : 1))
                            {   if  (  flag_v > 1
                                    && pluspacket * 3 + dstby < tcpcashe_count(l, reverse ? 0 : 1)
                                    )
                                    fprintf( stderr, "In file=%s record=%qu status=%08X %qu < %qu\n"
                                           , n_uniqt, n, status
                                           , dstby, tcpcashe_count(l, reverse ? 0 : 1)
                                           );
    ;       ;                   dstby = 0;
    ;       ;               }else dstby -= tcpcashe_count(l, reverse ? 0 : 1);
    ;       ;           }
    ;       ;           del_counter_index( &next.ws_ip_src
                                         , &next.ws_ip_dst
                                         , &next.ws_ip.sport
                                         , &next.ws_ip.dport
                                         );
    ;       ;       }else
                    {   tcpcashe_count(l, reverse ? 1 : 0) += srcby;
    ;       ;           tcpcashe_count(l, reverse ? 0 : 1) += dstby;
    ;       ;   }   }
    ;       ;   if  (eatoq(&next.ws_ip_phys.ethersrc) == eth)
                {   if  (status & REVERSE)
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu %s  srcR%s", n, proto, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_dst, dstcnt, srccnt, dstby, srcby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_src, dstcnt, srccnt);
    ;       ;       }else
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu %s  src %s", n, proto, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_dst, srccnt, dstcnt, srcby, dstby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_src, srccnt, dstcnt);
    ;       ;   }   }
    ;       ;   if  (eatoq(&next.ws_ip_phys.etherdst) == eth)
                {   if  (status & REVERSE)
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu %s  dstR%s", n, proto, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_src, srccnt, dstcnt, srcby, dstby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_dst, srccnt, dstcnt);
    ;       ;       }else
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu %s  dst %s", n, proto, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_src, dstcnt, srccnt, dstby, srcby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_dst, dstcnt, srccnt);
    ;       ;   }   }
    ;       }else if  (  status & FRAG_ONLY)
            {   srccnt = 1;
    ;       ;   dstcnt = 0;
    ;       ;   srcby = pluspacket * srccnt + ntohs(next.ws_ip_frag.currlen);
    ;       ;   dstby = 0;
    ;       ;   if  (eatoq(&next.ws_ip_phys.ethersrc) == eth)
                {   if  (status & REVERSE)
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu frag srcR%s", n, mc);
    ;       ;           uniqt("frag srcR");
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_dst, dstcnt, srccnt, dstby, srcby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_src, dstcnt, srccnt);
    ;       ;       }else
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu frag src %s", n, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_dst, srccnt, dstcnt, srcby, dstby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_src, srccnt, dstcnt);
    ;       ;   }   }
    ;       ;   if  (eatoq(&next.ws_ip_phys.etherdst) == eth)
                {   if  (status & REVERSE)
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu frag dstR%s", n, mc);
    ;       ;           uniqt("frag dstR");
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_src, srccnt, dstcnt, srcby, dstby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_dst, srccnt, dstcnt);
    ;       ;       }else
                    {   if  (flag_v > 1) fprintf(stderr, "%6qu frag dst %s", n, mc);
    ;       ;           if  (flag_t) do_sum( &next.ws_ip_src, dstcnt, srccnt, dstby, srcby);
    ;       ;           if  (flag_i) ip_registry(&next.ws_ip_dst, dstcnt, srccnt);
    }   }   }   }   }
                                                                               /* delete empty cashe */
    if  (flag_k > 2)
    {   int    i, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, ib;
        u_int32_t t0, t1, t2, t3, t4, t5, t6, t7, t8, t9, ta, tb;
        i = 0;
        for (i0 = 0; i0 < 256; i0++)
        { if ((t0 = tcpcashe_addr(0, i0)))
          { for (i1 = 0; i1 < 256; i1++)
            { if ((t1 = tcpcashe_addr(t0, i1)))
              { for (i2 = 0; i2 < 256; i2++)
                { if ((t2 = tcpcashe_addr(t1, i2)))
                  { for (i3 = 0; i3 < 256; i3++)
                    { if ((t3 = tcpcashe_addr(t2, i3)))
                      { for (i4 = 0; i4 < 256; i4++)
                        { if ((t4 = tcpcashe_addr(t3, i4)))
                          { for (i5 = 0; i5 < 256; i5++)
                            { if ((t5 = tcpcashe_addr(t4, i5)))
                              { for (i6 = 0; i6 < 256; i6++)
                                { if ((t6 = tcpcashe_addr(t5, i6)))
                                  { for (i7 = 0; i7 < 256; i7++)
                                    { if ((t7 = tcpcashe_addr(t6, i7)))
                                      { for (i8 = 0; i8 < 256; i8++)
                                        { if ((t8 = tcpcashe_addr(t7, i8)))
                                          { for (i9 = 0; i9 < 256; i9++)
                                            { if ((t9 = tcpcashe_addr(t8, i9)))
                                              { for (ia = 0; ia < 256; ia++)
                                                { if ((ta = tcpcashe_addr(t9, ia)))
                                                  { for (ib = 0; ib < 256; ib++)
                                                    { if ((tb = tcpcashe_addr(ta, ib)))
                                                      { if ( (tcpcashe_count(tb, 2) + timeout)
                                                           < globalast
                                                           )
                                                        { timeout_counter_index( i0, i1, i2, i3, i4, i5
                                                                               , i6, i7, i8, i9, ia, ib
                                                                               );
                                                        }else i++;
        } } } } } } } } } } } } } } } } } } } } } } } }
        if  (i == 0)
        {   if  (close(tcpcashe_addr_file))        err(EX_IOERR, "Can't close %s", tcpcashe_addr_name);
            if  (close(tcpcashe_count_file))      err(EX_IOERR, "Can't close %s", tcpcashe_count_name);
            if  (unlink(tcpcashe_addr_name))      err(EX_IOERR, "Can't unlink %s", tcpcashe_addr_name);
            if  (unlink(tcpcashe_count_name))    err(EX_IOERR, "Can't unlink %s", tcpcashe_count_name);
    }   }

                                                                      /* view IP addrs for given MAC */
    if  (flag_i)
    {   int i0, i1, i2, i3;
        bzero(&out, sizeof(out));
        if (ip_registered)
        { for (i0 = 0; i0 < 256; i0++)
          { if (ip_registered[i0])
            { for (i1 = 0; i1 < 256; i1++)
              { if (ip_registered[i0][i1])
                { for (i2 = 0; i2 < 256; i2++)
                  { if (ip_registered[i0][i1][i2])
                    { for (i3 = 0; i3 < 256; i3++)
                      { if  (   ip_registered[i0][i1][i2][i3].from
                            ||  ip_registered[i0][i1][i2][i3].to
                            )
                        {   if  (flag_b)
                            {   out.eth_addr = eth;
                        ;       out.iad[0] = i0;
                        ;       out.iad[1] = i1;
                        ;       out.iad[2] = i2;
                        ;       out.iad[3] = i3;
                        ;       out.i_record.sent = ip_registered[i0][i1][i2][i3].from;
                        ;       out.i_record.rcvd = ip_registered[i0][i1][i2][i3].to;
                        ;       if  (   sizeof(out)
                                    !=  write(fileno(stdout), &out, sizeof(out))
                                    )                  err(EX_IOERR, "%s: write stdout error", myname);
                        ;   }else
                            {   printf( "%u.%u.%u.%u\t%20qu\t%20qu\n"
                                       , i0, i1, i2, i3
                                       , ip_registered[i0][i1][i2][i3].from
                                       , ip_registered[i0][i1][i2][i3].to
                                       );
    }   } } } } } } } } }   }

                                                         /* view traffic in different account planes */
    if  (flag_t)
    {   bzero(&out, sizeof(out));
        if  (flag_b)
        {   for (l = 0; l < c_max; l++)
            {   if  (pcount[l] || bcount[l])
                {   out.eth_addr = eth;
                ;   out.t_record.category = l;
                ;   out.t_record.pcount = pcount[l];
                ;   out.t_record.bcount = bcount[l];
                ;   if  (   sizeof(out)
                        !=  write(fileno(stdout), &out, sizeof(out))
                        )                              err(EX_IOERR, "%s: write stdout error", myname);
            }   }
        }else if (flag_v)
        {   printf("category              packets                bytes avg pkt\n");
            for (l = 0; l < c_max; l++)
            {   if  (pcount[l] || bcount[l])
                {   printf( "   %5u %20qu %20qu  %6d\n"
                          , l, pcount[l], bcount[l]
                          , (int)(bcount[l]/pcount[l])
                          );
            }   }
        }else
        {   printf("category              packets                bytes\n");
            for (l = 0; l < c_max; l++)
            {   if  (pcount[l] || bcount[l])
                {   printf("   %5u %20qu %20qu\n", l, pcount[l], bcount[l]);
    }   }   }   }
    if  (flag_v) fprintf(stderr, "Max counter %u, max addr %u used\n", count_max_mon, addr_max_mon);
    exit(EX_OK);
}
